/*----- PROTECTED REGION ID(BacnetConf.cpp) ENABLED START -----*/
/* clang-format on */
//=============================================================================
//
// file :        BacnetConf.cpp
//
// description : C++ source for the BacnetConf class and its commands.
//               The class is derived from Device. It represents the
//               CORBA servant object which will be accessed from the
//               network. All commands which can be executed on the
//               BacnetConf are implemented in this file.
//
// project :     BacnetConf
//
// This file is part of Tango device class.
//
// Tango is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Tango is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Tango.  If not, see <http://www.gnu.org/licenses/>.
//
//
//
//=============================================================================
//                This file is generated by POGO
//        (Program Obviously used to Generate tango Object)
//=============================================================================


#include "BacnetConf.h"
#include "BacnetConfClass.h"
/* BACnet Stack API */
#include "bacnet/bactext.h"

/* clang-format off */
/*----- PROTECTED REGION END -----*/	//	BacnetConf.cpp

/**
 *  BacnetConf class description:
 *    Bacnet configurable device server
 */

//================================================================
//  The following table gives the correspondence
//  between command and method names.
//
//  Command name  |  Method name
//================================================================
//  State         |  Inherited (no method)
//  Status        |  Inherited (no method)
//  FindName      |  find_name
//  FindObjects   |  find_objects
//================================================================

//================================================================
//  Attributes managed is:
//================================================================
//================================================================

namespace BacnetConf_ns
{
/*----- PROTECTED REGION ID(BacnetConf::namespace_starting) ENABLED START -----*/
/* clang-format on */


struct StringUtils
{
	static std::map<std::string, std::string> domain_map;
	static std::map<Tango::DevState,std::string> tango_states;
	static void string_explode(std::string str, const std::string& separator, std::vector<std::string>& results);
	static void string_vector2map(const std::vector<std::string> &str, const std::string &separator, std::map<std::string,std::string> &results);
	static void replace_string(std::string& str, const std::string& from, const std::string& to);
	static std::string getFilename(const std::string &path, const std::string &dir, const std::string &name);
	static std::string fixName(const std::string &name);
	static std::string filename2attrname(const std::string &dir, const std::string &name);
};

std::map<std::string, std::string> StringUtils::domain_map;
void StringUtils::string_explode(std::string str, const std::string& separator, std::vector<std::string>& results)
{
	auto found = str.find_first_of(separator);
	while(found != std::string::npos)
	{
		if(found > 0)
		{
			results.push_back(str.substr(0,found));
		}
		str = str.substr(found+1);
		found = str.find_first_of(separator);
	}
	if(str.length() > 0) {
		results.push_back(str);
	}
}

void StringUtils::string_vector2map(const std::vector<std::string> &str, const std::string &separator, std::map<std::string,std::string> &results)
{
	for(const auto &it : str)
	{
		auto found_eq = it.find_first_of(separator);
		if(found_eq != std::string::npos && found_eq > 0)
		{
			results.insert(make_pair(it.substr(0,found_eq),it.substr(found_eq+1)));
		}
	}
}

void StringUtils::replace_string(std::string& str, const std::string& from, const std::string& to)
{
	if(from.empty())
		return;
	size_t start_pos = 0;
	while((start_pos = str.find(from, start_pos)) != std::string::npos)
	{
		str.replace(start_pos, from.length(), to);
		start_pos += to.length(); // In case 'to' contains 'from', like replacing 'x' with 'yx'
	}
}

std::string StringUtils::getFilename(const std::string &path, const std::string &dir, const std::string &name)
{
	std::string ret = path + '/' + dir + '/' + name;
	return ret;
}

std::string StringUtils::fixName(const std::string &name)
{
	std::string ret(name);
	ret.erase( std::remove( ret.begin(), ret.end(), '"' ), ret.end() );
	ret.erase( std::remove( ret.begin(), ret.end(), ' ' ), ret.end() );
	return ret;
}

std::string StringUtils::filename2attrname(const std::string &dir, const std::string &name)
{
	std::string ret = dir + '/' + name;
	//trim / at beginning
	ret.erase(0, ret.find_first_not_of('/'));
	//trim / at end
	ret.erase(ret.find_last_not_of('/') + 1);
	StringUtils::replace_string(ret, "/", "__");
	return ret;
}

/* clang-format off */
/*----- PROTECTED REGION END -----*/	//	BacnetConf::namespace_starting

//--------------------------------------------------------
/**
 *	Method      : BacnetConf::BacnetConf()
 * Description:  Constructors for a Tango device
 *                implementing the classBacnetConf
 */
//--------------------------------------------------------
BacnetConf::BacnetConf(Tango::DeviceClass *cl, std::string &s)
 : TANGO_BASE_CLASS(cl, s.c_str())
{
	/*----- PROTECTED REGION ID(BacnetConf::constructor_1) ENABLED START -----*/
	/* clang-format on */
	init_device();
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	BacnetConf::constructor_1
}
//--------------------------------------------------------
BacnetConf::BacnetConf(Tango::DeviceClass *cl, const char *s)
 : TANGO_BASE_CLASS(cl, s)
{
	/*----- PROTECTED REGION ID(BacnetConf::constructor_2) ENABLED START -----*/
	/* clang-format on */
	init_device();
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	BacnetConf::constructor_2
}
//--------------------------------------------------------
BacnetConf::BacnetConf(Tango::DeviceClass *cl, const char *s, const char *d)
 : TANGO_BASE_CLASS(cl, s, d)
{
	/*----- PROTECTED REGION ID(BacnetConf::constructor_3) ENABLED START -----*/
	/* clang-format on */
	init_device();
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	BacnetConf::constructor_3
}
//--------------------------------------------------------
BacnetConf::~BacnetConf()
{
	delete_device();
}

//--------------------------------------------------------
/**
 *	Method      : BacnetConf::delete_device()
 * Description:  will be called at device destruction or at init command
 */
//--------------------------------------------------------
void BacnetConf::delete_device()
{
	DEBUG_STREAM << "BacnetConf::delete_device() " << device_name << std::endl;
	/*----- PROTECTED REGION ID(BacnetConf::delete_device) ENABLED START -----*/
	/* clang-format on */
	//	Delete device allocated objects
	loop->abortflag = true;
	loop->signal();
	timespec time_to_sleep;
	time_to_sleep.tv_sec=0;
	time_to_sleep.tv_nsec=100000000L;//100ms
	nanosleep(&time_to_sleep, NULL);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	BacnetConf::delete_device
}

//--------------------------------------------------------
/**
 *	Method      : BacnetConf::init_device()
 * Description:  will be called at device initialization.
 */
//--------------------------------------------------------
void BacnetConf::init_device()
{
	DEBUG_STREAM << "BacnetConf::init_device() create device " << device_name << std::endl;
	/*----- PROTECTED REGION ID(BacnetConf::init_device_before) ENABLED START -----*/
	/* clang-format on */
	//	Initialization before get_device_property() call
	created_attr = false;
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	BacnetConf::init_device_before


	//	Get the device properties from database
	get_device_property();

	/*----- PROTECTED REGION ID(BacnetConf::init_device) ENABLED START -----*/
	/* clang-format on */
	//	Initialize device

	int ret;
	if(bACNET_BBMD_ADDRESS.length() > 0)
	{
		DEBUG_STREAM << __func__<<": exporting BACNET_BBMD_ADDRESS=" << bACNET_BBMD_ADDRESS;
		ret = setenv("BACNET_BBMD_ADDRESS", bACNET_BBMD_ADDRESS.c_str(), 1/*overwrite*/);
		if(ret < 0)
		{
			INFO_STREAM << __func__<<": ERROR exporting BACNET_BBMD_ADDRESS=" << bACNET_BBMD_ADDRESS << " err="<<errno<<"-"<<strerror(errno);
			//perror("setenv");
		}
	}
	if(bACNET_BBMD_PORT != 0)
	{
		DEBUG_STREAM << __func__<<": exporting BACNET_BBMD_PORT=" << bACNET_BBMD_PORT;
		std::stringstream bbmd_port;
		bbmd_port << bACNET_BBMD_PORT;
		ret = setenv("BACNET_BBMD_PORT", bbmd_port.str().c_str(), 1/*overwrite*/);
		if(ret < 0)
		{
			INFO_STREAM << __func__<<": ERROR exporting BACNET_BBMD_PORT=" << bACNET_BBMD_PORT << " err="<<errno<<"-"<<strerror(errno);
			//perror("setenv");
		}
	}
	if(bACNET_IP_PORT != 0)
	{
		DEBUG_STREAM << __func__<<": exporting BACNET_IP_PORT=" << bACNET_IP_PORT;
		std::stringstream bbmd_port;
		bbmd_port << bACNET_IP_PORT;
		ret = setenv("BACNET_IP_PORT", bbmd_port.str().c_str(), 1/*overwrite*/);
		if(ret < 0)
		{
			INFO_STREAM << __func__<<": ERROR exporting BACNET_IP_PORT=" << bACNET_IP_PORT << " err="<<errno<<"-"<<strerror(errno);
			//perror("setenv");
		}
	}

	if(!name_Map_File.empty())
	{
		ifstream infile (name_Map_File.c_str(), ios_base::in);
		std::string line;
		while (getline(infile, line, '\n'))
		{
			std::vector<std::string> v_line;
			StringUtils::string_explode(line,",",v_line);
			if(v_line.size() >= 4)
			{
				name_t obj;
				obj.s_device_instance = v_line[0];
				obj.s_object_type = v_line[1];
				obj.s_object_instance = v_line[2];
				obj.s_property=string("NO");
				if(v_line.size() >= 5)
				{
					if(v_line[4].find("NO") != std::string::npos)
					{
						continue; //no present value
					}
					obj.s_property = v_line[4];
				}
				name_obj_map.insert(make_pair(v_line[3],obj));

				object_t objs;
				objs.s_object_type = v_line[1];
				objs.s_object_instance = v_line[2];
				objs.s_object_name = v_line[3];
				objs.s_property = obj.s_property;
				std::string device(v_line[0]);
				objs.data_type = data_type_t::BAC_Real; //TODO: default
				if(v_line.size() == 6)
				{
					std::string s_data_type = v_line[5];
					try
					{
						objs.data_type = map_data_type.at(s_data_type);
					}
					catch(std::out_of_range& e)
					{
						
					}
				}
				dev_map_t::iterator it_dev = dev_name_map.find(device);
				if(it_dev == dev_name_map.end())
				{
					vector<object_t> v_objs;
					v_objs.push_back(objs);
					dev_name_map.insert(make_pair(device,v_objs));
				}
				else
				{
					it_dev->second.push_back(objs);
				}
			}
		}
		infile.close();
	}

	loop = new readthread(this);
	loop->start();
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	BacnetConf::init_device
}

//--------------------------------------------------------
/**
 *	Method      : BacnetConf::get_device_property()
 * Description:  Read database to initialize property data members.
 */
//--------------------------------------------------------
void BacnetConf::get_device_property()
{
	/*----- PROTECTED REGION ID(BacnetConf::get_device_property_before) ENABLED START -----*/
	/* clang-format on */
	//	Initialize property data members
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	BacnetConf::get_device_property_before


	//	Read device properties from database.
	Tango::DbData	dev_prop;
	dev_prop.push_back(Tango::DbDatum("Attr_config_prop"));
	dev_prop.push_back(Tango::DbDatum("Attr_config_file"));
	dev_prop.push_back(Tango::DbDatum("BACNET_BBMD_ADDRESS"));
	dev_prop.push_back(Tango::DbDatum("BACNET_IP_PORT"));
	dev_prop.push_back(Tango::DbDatum("BACNET_BBMD_PORT"));
	dev_prop.push_back(Tango::DbDatum("PollingPeriod"));
	dev_prop.push_back(Tango::DbDatum("Name_Map_File"));
	dev_prop.push_back(Tango::DbDatum("MaxReadPropertyMultiple"));

	//	is there at least one property to be read ?
	if (dev_prop.size()>0)
	{
		//	Call database and extract values
		if (Tango::Util::instance()->_UseDb==true)
			get_db_device()->get_property(dev_prop);

		//	get instance on BacnetConfClass to get class property
		Tango::DbDatum	def_prop, cl_prop;
		BacnetConfClass	*ds_class =
			(static_cast<BacnetConfClass *>(get_device_class()));
		int	i = -1;

		//	Try to initialize Attr_config_prop from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  attr_config_prop;
		else {
			//	Try to initialize Attr_config_prop from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  attr_config_prop;
		}
		//	And try to extract Attr_config_prop value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  attr_config_prop;

		//	Try to initialize Attr_config_file from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  attr_config_file;
		else {
			//	Try to initialize Attr_config_file from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  attr_config_file;
		}
		//	And try to extract Attr_config_file value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  attr_config_file;

		//	Try to initialize BACNET_BBMD_ADDRESS from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  bACNET_BBMD_ADDRESS;
		else {
			//	Try to initialize BACNET_BBMD_ADDRESS from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  bACNET_BBMD_ADDRESS;
		}
		//	And try to extract BACNET_BBMD_ADDRESS value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  bACNET_BBMD_ADDRESS;

		//	Try to initialize BACNET_IP_PORT from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  bACNET_IP_PORT;
		else {
			//	Try to initialize BACNET_IP_PORT from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  bACNET_IP_PORT;
		}
		//	And try to extract BACNET_IP_PORT value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  bACNET_IP_PORT;

		//	Try to initialize BACNET_BBMD_PORT from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  bACNET_BBMD_PORT;
		else {
			//	Try to initialize BACNET_BBMD_PORT from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  bACNET_BBMD_PORT;
		}
		//	And try to extract BACNET_BBMD_PORT value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  bACNET_BBMD_PORT;

		//	Try to initialize PollingPeriod from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  pollingPeriod;
		else {
			//	Try to initialize PollingPeriod from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  pollingPeriod;
		}
		//	And try to extract PollingPeriod value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  pollingPeriod;

		//	Try to initialize Name_Map_File from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  name_Map_File;
		else {
			//	Try to initialize Name_Map_File from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  name_Map_File;
		}
		//	And try to extract Name_Map_File value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  name_Map_File;

		//	Try to initialize MaxReadPropertyMultiple from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  maxReadPropertyMultiple;
		else {
			//	Try to initialize MaxReadPropertyMultiple from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  maxReadPropertyMultiple;
		}
		//	And try to extract MaxReadPropertyMultiple value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  maxReadPropertyMultiple;

	}

	/*----- PROTECTED REGION ID(BacnetConf::get_device_property_after) ENABLED START -----*/
	/* clang-format on */
	//	Check device property data members init
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	BacnetConf::get_device_property_after
}

//--------------------------------------------------------
/**
 *	Method      : BacnetConf::always_executed_hook()
 * Description:  method always executed before any command is executed
 */
//--------------------------------------------------------
void BacnetConf::always_executed_hook()
{
	DEBUG_STREAM << "BacnetConf::always_executed_hook()  " << device_name << std::endl;
	/*----- PROTECTED REGION ID(BacnetConf::always_executed_hook) ENABLED START -----*/
	/* clang-format on */
	//	code always executed before all requests
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	BacnetConf::always_executed_hook
}

//--------------------------------------------------------
/**
 *	Method      : BacnetConf::read_attr_hardware()
 * Description:  Hardware acquisition for attributes
 */
//--------------------------------------------------------
void BacnetConf::read_attr_hardware(TANGO_UNUSED(std::vector<long> &attr_list))
{
	DEBUG_STREAM << "BacnetConf::read_attr_hardware(std::vector<long> &attr_list) entering... " << std::endl;
	/*----- PROTECTED REGION ID(BacnetConf::read_attr_hardware) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	BacnetConf::read_attr_hardware
}


//--------------------------------------------------------
/**
 *	Read attribute dynScalarFloatRO related method
 *
 *
 *	Data type:	Tango::DevFloat
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void BacnetConf::read_dynScalarFloatRO(Tango::Attribute &attr)
{
	DEBUG_STREAM << "BacnetConf::read_dynScalarFloatRO(Tango::Attribute &attr) entering... " << std::endl;
	Tango::DevFloat	*att_value = get_dynScalarFloatRO_data_ptr(attr.get_name());
	/*----- PROTECTED REGION ID(BacnetConf::read_dynScalarFloatRO) ENABLED START -----*/
	/* clang-format on */
	attr_map_t::iterator it_attr;

	it_attr = attr_map.find(attr.get_name());
	if(it_attr == attr_map.end())
	{
		std::stringstream tmp;
		tmp << "Error looking for attribute " << attr.get_name();
		Tango::Except::throw_exception(
					(const char *) "INTERNAL_ERROR",
					tmp.str(),
					(const char *) __func__, Tango::ERR);
	}
	*att_value = it_attr->second.value_f;
	//	Set the attribute value
	attr.set_value(att_value);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	BacnetConf::read_dynScalarFloatRO
}
//--------------------------------------------------------
/**
 *	Read attribute dynScalarInt16RO related method
 *
 *
 *	Data type:	Tango::DevShort
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void BacnetConf::read_dynScalarInt16RO(Tango::Attribute &attr)
{
	DEBUG_STREAM << "BacnetConf::read_dynScalarInt16RO(Tango::Attribute &attr) entering... " << std::endl;
	Tango::DevShort	*att_value = get_dynScalarInt16RO_data_ptr(attr.get_name());
	/*----- PROTECTED REGION ID(BacnetConf::read_dynScalarInt16RO) ENABLED START -----*/
	/* clang-format on */
	attr_map_t::iterator it_attr;

	it_attr = attr_map.find(attr.get_name());
	if(it_attr == attr_map.end())
	{
		std::stringstream tmp;
		tmp << "Error looking for attribute " << attr.get_name();
		Tango::Except::throw_exception(
					(const char *) "INTERNAL_ERROR",
					tmp.str(),
					(const char *) __func__, Tango::ERR);
	}
	*att_value = it_attr->second.value_s;
	//	Set the attribute value
	attr.set_value(att_value);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	BacnetConf::read_dynScalarInt16RO
}
//--------------------------------------------------------
/**
 *	Read attribute dynScalarInt32RO related method
 *
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void BacnetConf::read_dynScalarInt32RO(Tango::Attribute &attr)
{
	DEBUG_STREAM << "BacnetConf::read_dynScalarInt32RO(Tango::Attribute &attr) entering... " << std::endl;
	Tango::DevLong	*att_value = get_dynScalarInt32RO_data_ptr(attr.get_name());
	/*----- PROTECTED REGION ID(BacnetConf::read_dynScalarInt32RO) ENABLED START -----*/
	/* clang-format on */
	attr_map_t::iterator it_attr;

	it_attr = attr_map.find(attr.get_name());
	if(it_attr == attr_map.end())
	{
		std::stringstream tmp;
		tmp << "Error looking for attribute " << attr.get_name();
		Tango::Except::throw_exception(
					(const char *) "INTERNAL_ERROR",
					tmp.str(),
					(const char *) __func__, Tango::ERR);
	}
	*att_value = it_attr->second.value_l;
	//	Set the attribute value
	attr.set_value(att_value);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	BacnetConf::read_dynScalarInt32RO
}
//--------------------------------------------------------
/**
 *	Read attribute dynScalarByteRO related method
 *
 *
 *	Data type:	Tango::DevUChar
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void BacnetConf::read_dynScalarByteRO(Tango::Attribute &attr)
{
	DEBUG_STREAM << "BacnetConf::read_dynScalarByteRO(Tango::Attribute &attr) entering... " << std::endl;
	Tango::DevUChar	*att_value = get_dynScalarByteRO_data_ptr(attr.get_name());
	/*----- PROTECTED REGION ID(BacnetConf::read_dynScalarByteRO) ENABLED START -----*/
	/* clang-format on */
	attr_map_t::iterator it_attr;

	it_attr = attr_map.find(attr.get_name());
	if(it_attr == attr_map.end())
	{
		std::stringstream tmp;
		tmp << "Error looking for attribute " << attr.get_name();
		Tango::Except::throw_exception(
					(const char *) "INTERNAL_ERROR",
					tmp.str(),
					(const char *) __func__, Tango::ERR);
	}
	*att_value = it_attr->second.value_uc;
	//	Set the attribute value
	attr.set_value(att_value);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	BacnetConf::read_dynScalarByteRO
}
//--------------------------------------------------------
/**
 *	Read attribute dynScalarBoolRO related method
 *
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void BacnetConf::read_dynScalarBoolRO(Tango::Attribute &attr)
{
	DEBUG_STREAM << "BacnetConf::read_dynScalarBoolRO(Tango::Attribute &attr) entering... " << std::endl;
	Tango::DevBoolean	*att_value = get_dynScalarBoolRO_data_ptr(attr.get_name());
	/*----- PROTECTED REGION ID(BacnetConf::read_dynScalarBoolRO) ENABLED START -----*/
	/* clang-format on */
	attr_map_t::iterator it_attr;

	it_attr = attr_map.find(attr.get_name());
	if(it_attr == attr_map.end())
	{
		std::stringstream tmp;
		tmp << "Error looking for attribute " << attr.get_name();
		Tango::Except::throw_exception(
					(const char *) "INTERNAL_ERROR",
					tmp.str(),
					(const char *) __func__, Tango::ERR);
	}
	*att_value = it_attr->second.value_b;
	//	Set the attribute value
	attr.set_value(att_value);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	BacnetConf::read_dynScalarBoolRO
}
//--------------------------------------------------------
/**
 *	Read attribute dynScalarDoubleRO related method
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void BacnetConf::read_dynScalarDoubleRO(Tango::Attribute &attr)
{
	DEBUG_STREAM << "BacnetConf::read_dynScalarDoubleRO(Tango::Attribute &attr) entering... " << std::endl;
	Tango::DevDouble	*att_value = get_dynScalarDoubleRO_data_ptr(attr.get_name());
	/*----- PROTECTED REGION ID(BacnetConf::read_dynScalarDoubleRO) ENABLED START -----*/
	/* clang-format on */
	attr_map_t::iterator it_attr;

	it_attr = attr_map.find(attr.get_name());
	if(it_attr == attr_map.end())
	{
		std::stringstream tmp;
		tmp << "Error looking for attribute " << attr.get_name();
		Tango::Except::throw_exception(
					(const char *) "INTERNAL_ERROR",
					tmp.str(),
					(const char *) __func__, Tango::ERR);
	}
	*att_value = it_attr->second.value_d;
	//	Set the attribute value
	attr.set_value(att_value);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	BacnetConf::read_dynScalarDoubleRO
}
//--------------------------------------------------------
/**
 *	Read attribute dynScalarUInt32RO related method
 *
 *
 *	Data type:	Tango::DevULong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void BacnetConf::read_dynScalarUInt32RO(Tango::Attribute &attr)
{
	DEBUG_STREAM << "BacnetConf::read_dynScalarUInt32RO(Tango::Attribute &attr) entering... " << std::endl;
	Tango::DevULong	*att_value = get_dynScalarUInt32RO_data_ptr(attr.get_name());
	/*----- PROTECTED REGION ID(BacnetConf::read_dynScalarUInt32RO) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(att_value);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	BacnetConf::read_dynScalarUInt32RO
}
//--------------------------------------------------------
/**
 *	Method      : BacnetConf::add_dynamic_attributes()
 * Description:  Create the dynamic attributes if any
 *                for specified device.
 */
//--------------------------------------------------------
void BacnetConf::add_dynamic_attributes()
{
	//	Example to add dynamic attribute:
	//	Copy inside the following protected area to create instance(s) at startup.
	//	add_dynScalarFloatRO_dynamic_attribute("MydynScalarFloatROAttribute");
	//	add_dynScalarInt16RO_dynamic_attribute("MydynScalarInt16ROAttribute");
	//	add_dynScalarInt32RO_dynamic_attribute("MydynScalarInt32ROAttribute");
	//	add_dynScalarByteRO_dynamic_attribute("MydynScalarByteROAttribute");
	//	add_dynScalarBoolRO_dynamic_attribute("MydynScalarBoolROAttribute");
	//	add_dynScalarDoubleRO_dynamic_attribute("MydynScalarDoubleROAttribute");
	//	add_dynScalarUInt32RO_dynamic_attribute("MydynScalarUInt32ROAttribute");

	/*----- PROTECTED REGION ID(BacnetConf::add_dynamic_attributes) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code to create and add dynamic attributes if any
	if (attr_config_prop.empty() == false)
	{
		DEBUG_STREAM << __FUNCTION__ << " going to create from device property...";
		create_dynamic_attributes(attr_config_prop);
	}

	if (attr_config_file.empty() == false)
	{
		DEBUG_STREAM << __FUNCTION__ << " going to create from file...";
		std::vector<std::string> attr_config;
		std::string line;
		attr_config.clear();
		ifstream infile (attr_config_file.c_str(), ios_base::in);
		while (getline(infile, line, '\n'))
		{
			attr_config.push_back (line);
		}
		infile.close();
		DEBUG_STREAM <<__FUNCTION__ << "Read " << attr_config.size() << " lines.\n";
		create_dynamic_attributes(attr_config);
	}
	/*if(attr_map.empty())
	{
		create_dynamic_attributes(attr_config_prop, true);
	}*/


	//if(done something)
	{
		for(attr_map_t::iterator attr_it=attr_map.begin(); attr_it != attr_map.end(); attr_it++)
		{
			if(attr_it->second.type == string(BAC_FLOAT))
				add_dynScalarFloatRO_dynamic_attribute(attr_it->second.name);
			else if(attr_it->second.type == string(BAC_DOUBLE))
				add_dynScalarDoubleRO_dynamic_attribute(attr_it->second.name);
			else if(attr_it->second.type == string(BAC_INT32))
				add_dynScalarInt32RO_dynamic_attribute(attr_it->second.name);
			else if(attr_it->second.type == string(BAC_UINT32))
				add_dynScalarUInt32RO_dynamic_attribute(attr_it->second.name);
			else if(attr_it->second.type == string(BAC_INT16))
				add_dynScalarInt16RO_dynamic_attribute(attr_it->second.name);
			else if(attr_it->second.type == string(BAC_BYTE))
				add_dynScalarByteRO_dynamic_attribute(attr_it->second.name);
			else if(attr_it->second.type == string(BAC_BOOL))
				add_dynScalarBoolRO_dynamic_attribute(attr_it->second.name);
		}
	}
	created_attr = true;
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	BacnetConf::add_dynamic_attributes
}

//--------------------------------------------------------
/**
 *	Command FindName related method
 * Description:  Looks for object-name in NameMap and return device and object
 *
 *	@param argin object-name
 *	@returns device-id,object-type,object-instance
 */
//--------------------------------------------------------
Tango::DevString BacnetConf::find_name(Tango::DevString argin)
{
	Tango::DevString argout;
	DEBUG_STREAM << "BacnetConf::FindName()  - " << device_name << std::endl;
	/*----- PROTECTED REGION ID(BacnetConf::find_name) ENABLED START -----*/
	/* clang-format on */

	//	Add your own code
	std::string object_name(argin);
	DEBUG_STREAM << "BacnetConf::FindName()  - looking for " << object_name << std::endl;
	name_map_t::iterator it_obj = name_obj_map.find(object_name);
	if(it_obj != name_obj_map.end())
	{
		stringstream s_object;
		s_object << it_obj->second.s_device_instance << ",(" << it_obj->second.s_object_type << "," << it_obj->second.s_object_instance << "),"<<object_name;
		argout = Tango::string_dup(s_object.str().c_str());
	}
	else
	{
		std::stringstream tmp;
		tmp << "object-name: " << object_name << " not found";
		Tango::Except::throw_exception(
					(const char *) "NOT_FOUND",
					tmp.str(),
					(const char *) __func__, Tango::ERR);
	}

	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	BacnetConf::find_name
	return argout;
}
//--------------------------------------------------------
/**
 *	Command FindObjects related method
 * Description:  Looks for device-id in NameMap and return list of objects
 *
 *	@param argin device-id
 *	@returns [0]: object-type,object-instance,object-name
 *	[1]: ...
 */
//--------------------------------------------------------
Tango::DevVarStringArray *BacnetConf::find_objects(Tango::DevString argin)
{
	Tango::DevVarStringArray *argout;
	DEBUG_STREAM << "BacnetConf::FindObjects()  - " << device_name << std::endl;
	/*----- PROTECTED REGION ID(BacnetConf::find_objects) ENABLED START -----*/
	/* clang-format on */

	//	Add your own code
	argout  = new Tango::DevVarStringArray();
	std::string device(argin);
	DEBUG_STREAM << "BacnetConf::FindObjects()  - looking for " << device << std::endl;
	dev_map_t::iterator it_dev = dev_name_map.find(device);
	if(it_dev != dev_name_map.end())
	{
		int i = 0;
		argout->length(it_dev->second.size());
		for (vector<object_t>::iterator it_obj= it_dev->second.begin(); it_obj != it_dev->second.end(); it_obj++) 
		{
			stringstream s_object;
			s_object << device << ",(" << it_obj->s_object_type << "," << it_obj->s_object_instance << ")," << it_obj->s_object_name;
			DEBUG_STREAM << "BacnetConf::FindObjects()  - looping " << s_object.str() << std::endl;
			(*argout)[i] = Tango::string_dup(s_object.str().c_str());
			i++;
		}
	}
	else
	{
		std::stringstream tmp;
		tmp << "device-instance: " << device << " not found";
		Tango::Except::throw_exception(
					(const char *) "NOT_FOUND",
					tmp.str(),
					(const char *) __func__, Tango::ERR);
	}

	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	BacnetConf::find_objects
	return argout;
}
//--------------------------------------------------------
/**
 *	Method      : BacnetConf::add_dynamic_commands()
 * Description:  Create the dynamic commands if any
 *                for specified device.
 */
//--------------------------------------------------------
void BacnetConf::add_dynamic_commands()
{
	/*----- PROTECTED REGION ID(BacnetConf::add_dynamic_commands) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code to create and add dynamic commands if any
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	BacnetConf::add_dynamic_commands
}

/*----- PROTECTED REGION ID(BacnetConf::namespace_ending) ENABLED START -----*/
/* clang-format on */
//	Additional Methods
/**
 *	Method      : BacnetConf::create_dynamic_attributes()
 *	Description : Create the dynamic attributes if any
 *	              for specified device.
 */
//--------------------------------------------------------
void BacnetConf::create_dynamic_attributes(std::vector<std::string> attr_config, bool create_all_objects)
{
	DEBUG_STREAM << __FUNCTION__ << " entering...";
	std::vector<std::string> filter;

	for (unsigned int i = 0;i < attr_config.size() && !create_all_objects;i++)
	{
		std::vector<std::string> configuration;
		StringUtils::string_explode(attr_config[i], ",", configuration);
		std::map<std::string,std::string> single_att_conf;
		StringUtils::string_vector2map(configuration,":",single_att_conf);
		DEBUG_STREAM << __func__<< ": looping configuration " << attr_config[i] << endl;
		if(configuration.size() >= 2)
		{
			timeval now;
			gettimeofday(&now, NULL);
			std::string name;
			std::string type;
			std::string s_device_instance;//e.g. 2107393
			std::string s_object_type; //e.g. analog-input, analog-output
			std::string s_object_instance; //e.g. 49
			std::string s_property; //e.g. present-value
			std::string s_object_name;
			attr_desc_t tmp;
			tmp.value_f=0.0;
			tmp.value_d=0.0;
			tmp.value_l=0;
			tmp.value_s=0;
			tmp.value_uc=0;
			tmp.value_b=false;
			tmp.size=1; //supporting only scalar attributes
			bool found_name=false;
			bool found_type=false;
			bool found_object_name=false;
			bool found_device_instance=false;
			bool found_object_type=false;
			bool found_object_instance=false;
			bool found_property=false;
			try
			{
				name = single_att_conf.at("\"Name\"");
				name.erase( std::remove( name.begin(), name.end(), '"' ), name.end() );
				tmp.name = name;
				found_name=true;
			}
			catch(const std::out_of_range& e)
			{
				std::stringstream err;
				err << "Configuration parsing error looking for Name: " << e.what();
				DEBUG_STREAM << __func__<< ": " << err.str() << endl;
				//continue;
			}
			try
			{
				type = single_att_conf.at("\"Type\"");
				type.erase( std::remove( type.begin(), type.end(), '"' ), type.end() );
				tmp.type = type;
				found_type=true;
			}
			catch(const std::out_of_range& e)
			{
				std::stringstream err;
				err << "Configuration parsing error looking for Type: " << e.what();
				DEBUG_STREAM << __func__<< ": " << err.str() << endl;
				//continue;
			}
			try
			{
				s_device_instance = single_att_conf.at("\"Device_instance\"");
				s_device_instance.erase( std::remove( s_device_instance.begin(), s_device_instance.end(), '"' ), s_device_instance.end() );
				tmp.s_device_instance = s_device_instance;
				found_device_instance=true;
			}
			catch(const std::out_of_range& e)
			{
				std::stringstream err;
				err << "Configuration parsing error looking for Device_instance: " << e.what();
				DEBUG_STREAM << __func__<< ": " << err.str() << endl;
				//continue;
			}			

			try
			{
				s_object_name = single_att_conf.at("\"Object_name\"");
				//s_object_name.erase( std::remove( s_object_name.begin(), s_object_name.end(), '"' ), s_object_name.end() );
				tmp.s_object_name = s_object_name;
				found_object_name = true;
			}
			catch(const std::out_of_range& e)
			{
				std::stringstream err;
				err << "Not found Object_name: " << e.what();
				DEBUG_STREAM << __func__<< ": " << err.str() << endl;
			}
			try
			{
				s_property = single_att_conf.at("\"Property\"");
				s_property.erase( std::remove( s_property.begin(), s_property.end(), '"' ), s_property.end() );
				tmp.s_property = s_property;
				uint32_t property=MAX_BACNET_PROPERTY_ID+1;
				int32_t property_index=-1;
				char property_id_s[64];
				int scan_count = sscanf(s_property.c_str(), "%[^[][%u]", &property_id_s,&property_index);
				if(scan_count > 0)
				{
					if (bactext_property_strtol(s_property.c_str(), &property) ==	false)
					{
						DEBUG_STREAM << __func__<< ": BAD property=" << s_property;
						continue;
					}
				}
				if (property > MAX_BACNET_PROPERTY_ID)
				{
					DEBUG_STREAM << __func__<< ": BAD property=" << s_property <<" - not greater than "<<MAX_BACNET_PROPERTY_ID;
					continue;
				}
				tmp.property = property;
				if (scan_count > 1 && property_index >=0)
				{
					//rpm_property->propertyArrayIndex = property_index;
				}
				tmp.property_index = property_index;
				found_property=true;
			}
			catch(const std::out_of_range& e)
			{
				std::stringstream err;
				err << "Configuration parsing error looking for Property: " << e.what();
				DEBUG_STREAM << __func__<< ": " << err.str() << endl;
			}

			try
			{
				s_object_type = single_att_conf.at("\"Object_type\"");
				s_object_type.erase( std::remove( s_object_type.begin(), s_object_type.end(), '"' ), s_object_type.end() );
				tmp.s_object_type = s_object_type;
				found_object_type=true;
			}
			catch(const std::out_of_range& e)
			{
				std::stringstream err;
				err << "Configuration parsing error looking for Object_type: " << e.what();
				DEBUG_STREAM << __func__<< ": " << err.str() << endl;
			}

			try
			{
				s_object_instance = single_att_conf.at("\"Object_instance\"");
				s_object_instance.erase( std::remove( s_object_instance.begin(), s_object_instance.end(), '"' ), s_object_instance.end() );
				tmp.s_object_instance = s_object_instance;
				found_object_instance=true;
			}
			catch(const std::out_of_range& e)
			{
				std::stringstream err;
				err << "Configuration parsing error looking for Object_instance: " << e.what();
				DEBUG_STREAM << __func__<< ": " << err.str() << endl;
			}
			if(found_device_instance && found_property && !found_object_instance && !found_object_type && !found_name && !found_object_type)
			{
				//all objects
				DEBUG_STREAM << __func__ << ":  going to export all objects for device " <<tmp.s_device_instance<< std::endl;
				//####### check device instance #######
				uint32_t Target_Device_Object_Instance = strtol(tmp.s_device_instance.c_str(), NULL, 0);
				if (Target_Device_Object_Instance > BACNET_MAX_INSTANCE)
				{
					INFO_STREAM << __func__ << ": BAD device-instance="<<Target_Device_Object_Instance<<" - not greater than "<<BACNET_MAX_INSTANCE;
					continue;
				}
				std::string device = tmp.s_device_instance;
				tmp.device_instance = Target_Device_Object_Instance;
				dev_map_t::iterator it_dev = dev_name_map.find(tmp.s_device_instance);
				if(it_dev != dev_name_map.end())
				{
					//argout->length(it_dev->second.size());
					for (vector<object_t>::iterator it_obj= it_dev->second.begin(); it_obj != it_dev->second.end(); it_obj++) 
					{
						stringstream s_object;
						s_object << device << ",(" << it_obj->s_object_type << "," << it_obj->s_object_instance << ")," << it_obj->s_object_name << "," << it_obj->s_property;
						DEBUG_STREAM << __func__ << ":  - looping " << s_object.str() << std::endl;
						name=it_obj->s_object_name;
						StringUtils::replace_string(name, "'", "_");
						StringUtils::replace_string(name, "\"", "");
						tmp.name=name;
						tmp.s_object_name = it_obj->s_object_name;
						tmp.s_object_type = it_obj->s_object_type;
						tmp.s_object_instance = it_obj->s_object_instance;
						if(tmp.s_property != it_obj->s_property)
						{
							DEBUG_STREAM << __func__ << ":  - skipping " << s_object.str() << std::endl;
						}
						//####### check object type #######
						uint32_t object_type;
						bool status =
							bactext_object_type_strtol(tmp.s_object_type.c_str(), &object_type);
						if (status == false)
						{
							DEBUG_STREAM << __func__<< ": BAD object-type=" << s_object_type << endl;
							continue;
						}
						if (object_type >= MAX_BACNET_OBJECT_TYPE)
						{
							DEBUG_STREAM << __func__<< ": BAD object-type=" << s_object_type <<" - not greater than "<<MAX_BACNET_OBJECT_TYPE;
							continue;
						}
						tmp.object_type = object_type;
						//####### check object instance #######
						uint32_t object_instance = strtol(tmp.s_object_instance.c_str(), NULL, 0);
						if (object_instance > BACNET_MAX_INSTANCE)
						{
							DEBUG_STREAM << __func__<< ": BAD object-instance=" << s_object_instance <<" - not greater than "<<BACNET_MAX_INSTANCE;
							continue;
						}
						tmp.object_instance = object_instance;
						//####### deduce type #######
						if(tmp.s_object_type.find("binary") != std::string::npos && (it_obj->data_type == data_type_t::BAC_Enumerated || it_obj->data_type == data_type_t::BAC_Boolean))
						{
							tmp.type = BAC_BOOL;
						}
						else if(/*tmp.s_object_type.find("binary") != std::string::npos &&*/ it_obj->data_type == data_type_t::BAC_Real)
						{
							tmp.type = BAC_FLOAT;
						}
						else if(/*tmp.s_object_type.find("binary") != std::string::npos &&*/ it_obj->data_type == data_type_t::BAC_Unsigned_Int)
						{
							tmp.type = BAC_UINT32;
						}
						//####### OK, insert in map #######
						attr_map.insert(make_pair(name, tmp));
					}
				}
				else
				{
					std::stringstream tmp;
					tmp << "device-instance: " << device << " not found";
					INFO_STREAM << __func__ << ": " << tmp.str();
				}
			}
			else if(found_name && found_type && found_object_name && found_property)
			{
				if(type != string(BAC_FLOAT) && type != string(BAC_INT32) && type != string(BAC_INT16) && type != string(BAC_DOUBLE))
				{
					DEBUG_STREAM << __func__<< ": " << "Type has to be in[FLOAT,INT32,INT16,DOUBLE]";
					found_type=false;
					continue;
				}
				name_map_t::iterator it_obj = name_obj_map.find(s_object_name);
				if(it_obj != name_obj_map.end())
				{
					tmp.s_device_instance = it_obj->second.s_device_instance;
					tmp.s_object_type = it_obj->second.s_object_type;
					tmp.s_object_instance = it_obj->second.s_object_instance;
					DEBUG_STREAM << __func__<< ": found object-name=" << s_object_name <<": " << tmp.s_device_instance << ",("<<tmp.s_object_type<<","<<tmp.s_object_instance<<")";
					//####### check device instance #######
					uint32_t Target_Device_Object_Instance = strtol(tmp.s_device_instance.c_str(), NULL, 0);
					if (Target_Device_Object_Instance > BACNET_MAX_INSTANCE)
					{
						INFO_STREAM << __func__ << ": BAD device-instance="<<Target_Device_Object_Instance<<" - not greater than "<<BACNET_MAX_INSTANCE;
						continue;
					}
					tmp.device_instance = Target_Device_Object_Instance;
					//####### check object type #######
					uint32_t object_type;
					bool status =
						bactext_object_type_strtol(tmp.s_object_type.c_str(), &object_type);
					if (status == false)
					{
						DEBUG_STREAM << __func__<< ": BAD object-type=" << s_object_type << endl;
						continue;
					}
					if (object_type >= MAX_BACNET_OBJECT_TYPE)
					{
						DEBUG_STREAM << __func__<< ": BAD object-type=" << s_object_type <<" - not greater than "<<MAX_BACNET_OBJECT_TYPE;
						continue;
					}
					tmp.object_type = object_type;
					//####### check object instance #######
					uint32_t object_instance = strtol(tmp.s_object_instance.c_str(), NULL, 0);
					if (object_instance > BACNET_MAX_INSTANCE)
					{
						DEBUG_STREAM << __func__<< ": BAD object-instance=" << s_object_instance <<" - not greater than "<<BACNET_MAX_INSTANCE;
						continue;
					}
					tmp.object_instance = object_instance;
					//####### OK, insert in map #######
					attr_map.insert(make_pair(name, tmp));
				}
				else
				{
					DEBUG_STREAM << __func__<< ": BAD object-name=" << s_object_name <<" - not found in name map";
					//continue;
				}
			}
			else if(found_name && found_type && found_device_instance && found_object_type && found_object_instance && found_property)
			{
				if(type != string(BAC_FLOAT) && type != string(BAC_INT32) && type != string(BAC_INT16) && type != string(BAC_DOUBLE))
				{
					DEBUG_STREAM << __func__<< ": " << "Type has to be in[FLOAT,INT32,INT16,DOUBLE]";
					found_type=false;
					continue;
				}
				//####### check device instance #######
				uint32_t Target_Device_Object_Instance = strtol(tmp.s_device_instance.c_str(), NULL, 0);
				if (Target_Device_Object_Instance > BACNET_MAX_INSTANCE)
				{
					INFO_STREAM << __func__ << ": BAD device-instance="<<Target_Device_Object_Instance<<" - not greater than "<<BACNET_MAX_INSTANCE;
					continue;
				}
				tmp.device_instance = Target_Device_Object_Instance;
				//####### check object type #######
				uint32_t object_type;
				bool status =
					bactext_object_type_strtol(tmp.s_object_type.c_str(), &object_type);
				if (status == false)
				{
					DEBUG_STREAM << __func__<< ": BAD object-type=" << s_object_type << endl;
					continue;
				}
				if (object_type >= MAX_BACNET_OBJECT_TYPE)
				{
					DEBUG_STREAM << __func__<< ": BAD object-type=" << s_object_type <<" - not greater than "<<MAX_BACNET_OBJECT_TYPE;
					continue;
				}
				tmp.object_type = object_type;
				//####### check object instance #######
				uint32_t object_instance = strtol(tmp.s_object_instance.c_str(), NULL, 0);
				if (object_instance > BACNET_MAX_INSTANCE)
				{
					DEBUG_STREAM << __func__<< ": BAD object-instance=" << s_object_instance <<" - not greater than "<<BACNET_MAX_INSTANCE;
					continue;
				}
				tmp.object_instance = object_instance;
				//####### OK, insert in map #######
				attr_map.insert(make_pair(name, tmp));
			}
		}
	}

	//DEBUG_STREAM << __FUNCTION__ << " exiting...";
}
/* clang-format off */
/*----- PROTECTED REGION END -----*/	//	BacnetConf::namespace_ending
} //	namespace
