//
// file :		 readthread.cpp
//
// description :  thread
//
//
// copyleft :   Sincrotrone Trieste S.C.p.A.
//			  Strada Statale 14 - km 163,5 in AREA Science Park
//			  34012 Basovizza, Trieste ITALY
//

#include "readthread.h"
#include "BacnetConf.h"


/* BACnet Stack defines - first */
#include "bacnet/bacdef.h"
/* BACnet Stack API */
#include "bacnet/bactext.h"
#include "bacnet/bacerror.h"
#include "bacnet/iam.h"
#include "bacnet/arf.h"
#include "bacnet/npdu.h"
#include "bacnet/apdu.h"
#include "bacnet/rpm.h"
#include "bacnet/whois.h"
#include "bacnet/version.h"
/* some demo stuff needed */
#include "bacnet/basic/binding/address.h"
#include "bacnet/basic/object/device.h"
#include "bacnet/basic/sys/filename.h"
#include "bacnet/basic/services.h"
#include "bacnet/basic/tsm/tsm.h"
#include "bacnet/datalink/datalink.h"
#include "bacnet/datalink/dlenv.h"
#include "bacport.h"
#include "bacnet/basic/sys/keylist.h" //typedef KEYLIST_TYPE *OS_Keylist
#include "bacnet/property.h"//:bool property_list_writable_member(


namespace BacnetConf_ns
{

double dtdiff(timeval t1, timeval t2)
{
	double dt1 = t1.tv_sec + (double)t1.tv_usec/1000000.0;
	double dt2 = t2.tv_sec + (double)t2.tv_usec/1000000.0;
	return dt1-dt2;
}

template<typename T> void convertM(const BACNET_OBJECT_PROPERTY_VALUE *object_value, T &result)
{
	const BACNET_APPLICATION_DATA_VALUE *value;
	BACNET_PROPERTY_ID property = PROP_ALL;
	BACNET_OBJECT_TYPE object_type = MAX_BACNET_OBJECT_TYPE;
	//cout << endl<<__func__<<": entering"<<endl;
	/*if(!object_value)
		cout << __func__<<": !object_value"<<endl;
	if(object_value && !(object_value->value))
		cout << __func__<<": !object_value->value"<<endl;*/

	if (object_value && object_value->value) {
		//cout << __func__<<": object_value && object_value->value tag="<<(int)object_value->value->tag<<endl;
		value = object_value->value;
		property = object_value->object_property;
		object_type = object_value->object_type;
		switch (value->tag) {
#if defined(BACAPP_NULL)
			case BACNET_APPLICATION_TAG_NULL:
				//ret_val = bacapp_snprintf_null(str, str_len);
				break;
#endif
			case BACNET_APPLICATION_TAG_BOOLEAN:
				result = static_cast<T>(value->type.Boolean);
				break;
			case BACNET_APPLICATION_TAG_UNSIGNED_INT:
				result = static_cast<T>(value->type.Unsigned_Int);
				break;
			case BACNET_APPLICATION_TAG_SIGNED_INT:
				result = static_cast<T>(value->type.Signed_Int);
				break;
			case BACNET_APPLICATION_TAG_REAL:
				result = static_cast<T>(value->type.Real);
				break;
			case BACNET_APPLICATION_TAG_DOUBLE:
				result = static_cast<T>(value->type.Double);
				break;
#if 0//defined(BACAPP_OCTET_STRING)
			case BACNET_APPLICATION_TAG_OCTET_STRING:
				len = octetstring_length(&value->type.Octet_String);
				if (len > 0) {
					const uint8_t *octet_str;
					octet_str = octetstring_value(
						(BACNET_OCTET_STRING *)&value->type.Octet_String);
					for (i = 0; i < len; i++) {
						slen =
							bacapp_snprintf(str, str_len, "%02X", *octet_str);
						ret_val += bacapp_snprintf_shift(slen, &str, &str_len);
						octet_str++;
					}
				}
				break;
#endif
#if 0//defined(BACAPP_CHARACTER_STRING)
			case BACNET_APPLICATION_TAG_CHARACTER_STRING:
				len = characterstring_length(&value->type.Character_String);
				char_str = characterstring_value(&value->type.Character_String);
				slen = bacapp_snprintf(str, str_len, "\"");
				ret_val += bacapp_snprintf_shift(slen, &str, &str_len);
#if (__STDC_VERSION__ >= 199901L) && defined(__STDC_ISO_10646__)
				if (characterstring_encoding(&value->type.Character_String) ==
					CHARACTER_UTF8) {
					while (len > 0) {
						wclen = mbtowc(&wc, char_str, MB_CUR_MAX);
						if (wclen == -1) {
							/* Encoding error, reset state: */
							mbtowc(NULL, NULL, MB_CUR_MAX);
							/* After handling an invalid byte,
							   retry with the next one. */
							wclen = 1;
							wc = L'?';
						} else {
							if (!iswprint(wc)) {
								wc = L'.';
							}
						}
						/* For portability, cast wchar_t to wint_t */
						slen = bacapp_snprintf(str, str_len, "%lc", (wint_t)wc);
						ret_val += bacapp_snprintf_shift(slen, &str, &str_len);
						if (len > wclen) {
							len -= wclen;
							char_str += wclen;
						} else {
							len = 0;
						}
					}
				} else
#endif
				{
					for (i = 0; i < len; i++) {
						if (isprint(*((const unsigned char *)char_str))) {
							slen =
								bacapp_snprintf(str, str_len, "%c", *char_str);
						} else {
							slen = bacapp_snprintf(str, str_len, "%c", '.');
						}
						ret_val += bacapp_snprintf_shift(slen, &str, &str_len);
						char_str++;
					}
				}
				slen = bacapp_snprintf(str, str_len, "\"");
				ret_val += slen;
				break;
#endif
			case BACNET_APPLICATION_TAG_ENUMERATED:
			{
				/*ret_val = bacapp_snprintf_enumerated(
					str, str_len, object_type, property,
					value->type.Enumerated);*/
				uint32_t value_enum = value->type.Enumerated;
				switch (property) {
#if 0
					case PROP_PROPERTY_LIST:
						ret_val = bacapp_snprintf_property_identifier(str, str_len, value_enum);
						break;
					case PROP_OBJECT_TYPE:
						if (value_enum <= BACNET_OBJECT_TYPE_LAST) {
							ret_val = bacapp_snprintf(
								str, str_len, "%s", bactext_object_type_name(value_enum));
						} else if (value_enum <= BACNET_OBJECT_TYPE_RESERVED_MAX) {
							ret_val = bacapp_snprintf(
								str, str_len, "reserved %lu", (unsigned long)value_enum);
						} else {
							ret_val = bacapp_snprintf(
								str, str_len, "proprietary %lu", (unsigned long)value_enum);
						}
						break;
					case PROP_EVENT_STATE:
						ret_val = bacapp_snprintf(
							str, str_len, "%s", bactext_event_state_name(value_enum));
						break;
					case PROP_UNITS:
						if (bactext_engineering_unit_name_proprietary((unsigned)value_enum)) {
							ret_val = bacapp_snprintf(
								str, str_len, "proprietary %lu", (unsigned long)value_enum);
						} else {
							ret_val = bacapp_snprintf(
								str, str_len, "%s", bactext_engineering_unit_name(value_enum));
						}
						break;
					case PROP_POLARITY:
						ret_val = bacapp_snprintf(
							str, str_len, "%s", bactext_binary_polarity_name(value_enum));
						break;
#endif
					case PROP_PRESENT_VALUE:
					case PROP_RELINQUISH_DEFAULT:
					{
						switch (object_type) {
							case OBJECT_BINARY_INPUT:
							case OBJECT_BINARY_OUTPUT:
							case OBJECT_BINARY_VALUE:
							{
								std::string enum_label= string(bactext_binary_present_value_name(value_enum));
								//result = enum_label == string("active");
								result = static_cast<T>(value_enum);
								//cout<<__func__<<": enum_label="<<enum_label<<" -> result="<<(int)result<<endl;
								break;
							}
#if 0
							case OBJECT_BINARY_LIGHTING_OUTPUT:
								ret_val = bacapp_snprintf(
									str, str_len, "%s",
									bactext_binary_lighting_pv_name(value_enum));
								break;
#endif
							default:
								cout<<__func__<<": BACNET_APPLICATION_TAG_ENUMERATED, PROP_PRESENT_VALUE, Unknown Object type -> Value=" <<	value_enum << endl;
								break;
						}
						break;
					}
#if 0
					case PROP_RELIABILITY:
						ret_val = bacapp_snprintf(
							str, str_len, "%s", bactext_reliability_name(value_enum));
						break;
					case PROP_SYSTEM_STATUS:
						ret_val = bacapp_snprintf(
							str, str_len, "%s", bactext_device_status_name(value_enum));
						break;
					case PROP_SEGMENTATION_SUPPORTED:
						ret_val = bacapp_snprintf(
							str, str_len, "%s", bactext_segmentation_name(value_enum));
						break;
					case PROP_NODE_TYPE:
						ret_val = bacapp_snprintf(
							str, str_len, "%s", bactext_node_type_name(value_enum));
						break;
					case PROP_TRANSITION:
						ret_val = bacapp_snprintf(
							str, str_len, "%s", bactext_lighting_transition(value_enum));
						break;
					case PROP_IN_PROGRESS:
						ret_val = bacapp_snprintf(
							str, str_len, "%s", bactext_lighting_in_progress(value_enum));
						break;
#endif
					default:
						cout<<__func__<<": BACNET_APPLICATION_TAG_ENUMERATED, Unknown property -> Value=" <<value_enum << endl;
						break;
				}



				break;
			}
#if 0//defined(BACAPP_DATE)
			case BACNET_APPLICATION_TAG_DATE:
				ret_val = bacapp_snprintf_date(str, str_len, &value->type.Date);
				break;
#endif
#if 0//defined(BACAPP_TIME)
			case BACNET_APPLICATION_TAG_TIME:
				ret_val = bacapp_snprintf_time(str, str_len, &value->type.Time);
				break;
#endif


			case BACNET_APPLICATION_TAG_EMPTYLIST:
				//ret_val = bacapp_snprintf(str, str_len, "{}");
				break;
			default:
				cout << __func__<<": NOT handled tag="<<(int)value->tag<<endl;
				//ret_val = bacapp_snprintf(
				 //   str, str_len, "UnknownType(tag=%d)", value->tag);
				break;
		}
	}

	if (false) {
		stringstream err_mesg;
		//err_mesg << "Failed conversion of " << data;
		Tango::Except::throw_exception( "",
				err_mesg.str(),
				"convert()");
	}
}

template<typename T> void convertM(BACNET_OBJECT_TYPE object_type,
	uint32_t object_instance,
	BACNET_PROPERTY_REFERENCE *rpm_property, T &result)
{
	BACNET_OBJECT_PROPERTY_VALUE object_value;
	BACNET_APPLICATION_DATA_VALUE *value, *old_value;
	value = rpm_property->value;
	if (value == NULL)
	{
		std::stringstream tmp;
		tmp << "BACnet Error: " << bactext_error_class_name((int)rpm_property->error.error_class) << ": " << bactext_error_code_name((int)rpm_property->error.error_code);
		//cout << tmp.str() << endl;
		//TODO: throw exception
	}
	object_value.object_type = object_type;
	object_value.object_instance = object_instance;
	if ((value != NULL) && (value->next != NULL))
	{
		cout << __func__ << ": TODO ARRAY" << endl;
	}
	//while (value != NULL)//TODO: ARRAY!!!
	if (value != NULL)
	{
		object_value.object_property = rpm_property->propertyIdentifier;
		object_value.array_index = rpm_property->propertyArrayIndex;
		object_value.value = value;
		convertM(&object_value, result);//TODO: ARRAY!!!
		return;
	}
}

/* buffer used for receive */
/*static*/ uint8_t Rx_Buf[MAX_MPDU] = { 0 };

/* global variables used in this file */

/* needed to filter incoming messages + needed for return value of main application */
read_data_map_t read_access_data_map;

std::string print_bacnet_address(BACNET_ADDRESS *src)
{
	uint8_t i = 0;
	std::stringstream src_add;
	src_add << "MAC=";
	for (i = 0; i < MAX_MAC_LEN && i < src->mac_len; i++)
	{
		src_add <<hex<< (int)src->mac[i]<<dec;
		if(i < src->mac_len-1)
		{
			src_add << ".";
		}
	}
	src_add << " NET="<<src->net;
	src_add << " ADR=";
	for (i = 0; i < MAX_MAC_LEN && i < src->len; i++)
	{
		src_add <<hex<< (int)src->adr[i]<<dec;
		if(i < src->len-1)
		{
			src_add << ".";
		}
	}
	return src_add.str();
}

void MyErrorHandler(
	BACNET_ADDRESS *src,
	uint8_t invoke_id,
	BACNET_ERROR_CLASS error_class,
	BACNET_ERROR_CODE error_code)
{	
	cout << __func__ << ": entering invoke_id="<<(int)invoke_id<< " src:"<<print_bacnet_address(src)<< endl;

	read_data_map_t::iterator it;
	for(it=read_access_data_map.begin(); it!=read_access_data_map.end(); it++)
	{
		vector<read_data_t>::iterator it2;
		for(it2=it->second.begin(); it2 != it->second.end(); it2++)
		{
			if (address_match(&it2->Target_Address, src) &&
			(invoke_id == it2->Request_Invoke_ID))
			{
				it2->Error_Detected = true;
				it2->error_desc = std::string("BACnet error: ") + std::string(bactext_error_class_name((int)error_class)) + string("-") + std::string(bactext_error_code_name((int)error_code));
				cout << __func__ << ": " << it2->error_desc<<endl;
				break;
			}
		}
		if(it2 != it->second.end()-1)
			break;
	}
}

void MyAbortHandler(
	BACNET_ADDRESS *src, uint8_t invoke_id, uint8_t abort_reason, bool server)
{
	cout << __func__ << ": entering invoke_id="<<(int)invoke_id<< " src:"<<print_bacnet_address(src)<< endl;
	(void)server;
	read_data_map_t::iterator it;
	for(it=read_access_data_map.begin(); it!=read_access_data_map.end(); it++)
	{
		vector<read_data_t>::iterator it2;
		for(it2=it->second.begin(); it2 != it->second.end(); it2++)
		{
			if (address_match(&it2->Target_Address, src) &&
			(invoke_id == it2->Request_Invoke_ID))
			{
				it2->Error_Detected = true;
				it2->error_desc = std::string("BACnet Abort: ") + std::string(bactext_abort_reason_name((int)abort_reason));
				cout << __func__ << ": " << it2->error_desc<<endl;
				break;
			}
		}
		if(it2 != it->second.end()-1)
			break;
	}
}

void MyRejectHandler(BACNET_ADDRESS *src, uint8_t invoke_id, uint8_t reject_reason)
{
	cout << __func__ << ": entering invoke_id="<<(int)invoke_id<< " src:"<<print_bacnet_address(src)<< endl;
	/* FIXME: verify src and invoke id */
	read_data_map_t::iterator it;
	for(it=read_access_data_map.begin(); it!=read_access_data_map.end(); it++)
	{
		vector<read_data_t>::iterator it2;
		for(it2=it->second.begin(); it2 != it->second.end(); it2++)
		{
			if (address_match(&it2->Target_Address, src) &&
			(invoke_id == it2->Request_Invoke_ID))
			{
				it2->Error_Detected = true;
				it2->error_desc = std::string("BACnet Reject: ") + std::string(bactext_reject_reason_name((int)reject_reason));
				cout << __func__ << ": " << it2->error_desc<<endl;
				break;
			}
		}
		if(it2 != it->second.end()-1)
			break;
	}
}

//from apps/epics/main.c

void MyReadPropertyAckHandler(
	uint8_t *service_request,
	uint16_t service_len,
	BACNET_ADDRESS *src,
	BACNET_CONFIRMED_SERVICE_ACK_DATA *service_data)
{
	//cout << __func__ << ": entering invoke_id="<<(int)service_data->invoke_id<< " src:"<<print_bacnet_address(src)<< endl;

	int len = 0;
	BACNET_READ_ACCESS_DATA *rp_data;

	read_data_map_t::iterator it;
	for(it=read_access_data_map.begin(); it!=read_access_data_map.end(); it++)
	{
		vector<read_data_t>::iterator it2;
		for(it2=it->second.begin(); it2 != it->second.end(); it2++)
		{
			if (address_match(&it2->Target_Address, src) &&
			(service_data->invoke_id == it2->Request_Invoke_ID))
			{
				rp_data = static_cast<BACNET_READ_ACCESS_DATA*>(calloc(1, sizeof(BACNET_READ_ACCESS_DATA)));
				if (rp_data) {
					len = rp_ack_fully_decode_service_request(
						service_request, service_len, rp_data);
				}
				if (len > 0) {
					memmove(
						&it2->Read_Property_Multiple_Data.service_data, service_data,
						sizeof(BACNET_CONFIRMED_SERVICE_ACK_DATA));
					it2->Read_Property_Multiple_Data.rpm_data = rp_data;
					it2->Read_Property_Multiple_Data.new_data = true;
				} else {
					if (len < 0) { /* Eg, failed due to no segmentation */
						cout << __func__ << ": Error memmove len < 0" << endl;
						it2->Error_Detected = true;
					}
					free(rp_data);
				}
				break;
			}
		}
		if(it2 != it->second.end()-1)
			break;
	}
}


void MyReadPropertyMultipleAckHandler(
	uint8_t *service_request,
	uint16_t service_len,
	BACNET_ADDRESS *src,
	BACNET_CONFIRMED_SERVICE_ACK_DATA *service_data)
{
	//cout << __func__ << ": entering invoke_id="<<(int)service_data->invoke_id << " sequence_number="<<(int)service_data->sequence_number <<" segmented_message="<<(service_data->segmented_message ? "TRUE" : "FALSE") << " more_follows="<<(service_data->more_follows ? "TRUE" : "FALSE") << " from "<<print_bacnet_address(src)<< endl;
	int len = 0;
	bool found=false;
	BACNET_READ_ACCESS_DATA *rpm_data;

	read_data_map_t::iterator it;
	for(it=read_access_data_map.begin(); it!=read_access_data_map.end(); it++)
	{
		vector<read_data_t>::iterator it2;
		for(it2=it->second.begin(); it2 != it->second.end(); it2++)
		{
			//cout << __func__ << ": looping invoke_id="<<(int)it2->Request_Invoke_ID << " from "<<print_bacnet_address(&it2->Target_Address)<< endl;
			if (address_match(&it2->Target_Address, src) &&
			(service_data->invoke_id == it2->Request_Invoke_ID))
			{
				found=true;
				rpm_data = static_cast<BACNET_READ_ACCESS_DATA*>(calloc(1, sizeof(BACNET_READ_ACCESS_DATA)));
				if (rpm_data)
				{
					len = rpm_ack_decode_service_request(
						service_request, service_len, rpm_data);
				}
				//cout << __func__ << ": FOUND invoke_id="<<(int)it2->Request_Invoke_ID << " decoded len="<<len<<endl;
				if (len > 0)
				{
					memmove(
						&it2->Read_Property_Multiple_Data.service_data, service_data,
						sizeof(BACNET_CONFIRMED_SERVICE_ACK_DATA));
					it2->Read_Property_Multiple_Data.rpm_data = rpm_data;
					it2->Read_Property_Multiple_Data.new_data = true;
					/* Will process and free the RPM data later */
				}
				else
				{
					if (len < 0)
					{ /* Eg, failed due to no segmentation */
						cout << __func__ << ": Error memmove len < 0..." << endl;
						it2->Error_Detected = true;
					}
					rpm_data = rpm_data_free(rpm_data);
					free(rpm_data);
				}
				break;
			}
		}
		if(it2 != it->second.end()-1)
			break;
	}
	if(!found)
	{
		if (tsm_invoke_id_free(service_data->invoke_id))
		{
			//cout << __func__ << ": invoke_id="<<(int)service_data->invoke_id << " is free"<<endl;
		}
		else
		{
			//cout << __func__ << ": invoke_id="<<(int)service_data->invoke_id << " is NOT free"<<endl;
			if (tsm_invoke_id_failed(service_data->invoke_id))
			{
				cout << __func__ << ": invoke_id="<<(int)service_data->invoke_id << " -> Error: TSM Timeout!"<<endl;
				tsm_free_invoke_id(service_data->invoke_id);
			}
		}
	}
}

void readthread::Init_Service_Handlers(void)
{
	Device_Init(NULL);
	/* we need to handle who-is
	   to support dynamic device binding to us */
	apdu_set_unconfirmed_handler(SERVICE_UNCONFIRMED_WHO_IS, handler_who_is);
	/* handle i-am to support binding to other devices */
	apdu_set_unconfirmed_handler(SERVICE_UNCONFIRMED_I_AM, handler_i_am_bind);
	/* set the handler for all the services we don't implement
	   It is required to send the proper reject message... */
	apdu_set_unrecognized_service_handler_handler(handler_unrecognized_service);
	/* we must implement read property - it's required! */
	apdu_set_confirmed_handler(
		SERVICE_CONFIRMED_READ_PROPERTY, handler_read_property);
	/* handle the data coming back from confirmed requests */
	apdu_set_confirmed_ack_handler(
		SERVICE_CONFIRMED_READ_PROPERTY, MyReadPropertyAckHandler);
	apdu_set_confirmed_ack_handler(
		//SERVICE_CONFIRMED_READ_PROP_MULTIPLE, reinterpret_cast<confirmed_ack_function>(&readthread::MyReadPropertyMultipleAckHandler));
		SERVICE_CONFIRMED_READ_PROP_MULTIPLE, MyReadPropertyMultipleAckHandler);
	/* handle any errors coming back */
	apdu_set_error_handler(SERVICE_CONFIRMED_READ_PROPERTY, MyErrorHandler);
	apdu_set_abort_handler(MyAbortHandler);
	//apdu_set_reject_handler(dynamic_cast<reject_function>(&readthread::MyRejectHandler));
	apdu_set_reject_handler(MyRejectHandler);
}

void readthread::cleanup(void)
{
	cout << __func__ << ": entering..." << endl;
	BACNET_READ_ACCESS_DATA *rpm_object;
	BACNET_READ_ACCESS_DATA *old_rpm_object;
	BACNET_PROPERTY_REFERENCE *rpm_property;
	BACNET_PROPERTY_REFERENCE *old_rpm_property;

	for(read_data_map_t::iterator it=read_access_data_map.begin(); it!=read_access_data_map.end(); it++)
	{
		vector<read_data_t>::iterator it2;
		for(it2=it->second.begin(); it2 != it->second.end(); it2++)
		{
			rpm_object = it2->Read_Access_Data;
			old_rpm_object = rpm_object;
			while (rpm_object) {
				rpm_property = rpm_object->listOfProperties;
				while (rpm_property) {
					old_rpm_property = rpm_property;
					rpm_property = rpm_property->next;
					free(old_rpm_property);
				}
				old_rpm_object = rpm_object;
				rpm_object = rpm_object->next;
				free(old_rpm_object);
			}
		}
	}
}


/** Print out the value(s) for one Property.
 * This function may be called repeatedly for one property if we are walking
 * through a list (Using_Walked_List is True) to show just one value of the
 * array per call.
 *
 * @param object_type [in] The BACnet Object type of this object.
 * @param object_instance [in] The ID number for this object.
 * @param rpm_property [in] Points to structure holding the Property,
 *						  Value, and Error information.
 * @param result [out]
 */
template<typename T>  void ConvertPropertyData(
	BACNET_OBJECT_TYPE object_type,
	uint32_t object_instance,
	BACNET_PROPERTY_REFERENCE *rpm_property,
	T &result)
{
	//cout<<__func__<<": entering.."<<endl;
	BACNET_OBJECT_PROPERTY_VALUE object_value; /* for bacapp printing */
	BACNET_APPLICATION_DATA_VALUE *value, *old_value;
	bool print_brace = false;
	KEY object_list_element;
/* When we have to walk through an array of things, like ObjectIDs or
 * Subordinate_Annotations, one RP call at a time, use these for indexing.
 */
	uint32_t Walked_List_Length = 0;
	uint32_t Walked_List_Index = 0;
/* We get the length of the object list,
   and then get the objects one at a time */
	uint32_t Object_List_Length = 0;
	int32_t Object_List_Index = 0;
	/* object that we are currently printing */
	OS_Keylist Object_List;

	if (rpm_property == NULL) {
		//fprintf(stdout, "	-- Null Property data \n");
		//TODO: throw
		return;
	}
	value = rpm_property->value;
	if (value == NULL) {
		/* Then we print the error information */
		fprintf(
			stdout, "?  -- BACnet Error: %s: %s\n",
			bactext_error_class_name((int)rpm_property->error.error_class),
			bactext_error_code_name((int)rpm_property->error.error_code));
		//TODO: throw
		return;
	}
	object_value.object_type = object_type;
	object_value.object_instance = object_instance;
	if ((value != NULL) && (value->next != NULL)) {
		/* Then this is an array of values.
		 * But are we showing Values?  We (VTS3) want ? instead of {?,?} to show
		 * up. */
		switch (rpm_property->propertyIdentifier) {
				/* Screen the Properties that can be arrays or Sequences */
			case PROP_PRESENT_VALUE:
			case PROP_PRIORITY_ARRAY:
				if (object_type == OBJECT_DATETIME_VALUE) {
					break; /* A special case - no braces for this pair */
				}
				BACNET_STACK_FALLTHROUGH();
			default:
				/* Normal array: open brace */
				//fprintf(stdout, "{ ");
				print_brace = true; /* remember to close it */
				break;
		}
	}

	if (true/*!Using_Walked_List*/) {
		Walked_List_Index = Walked_List_Length = 0; /* In case we need this. */
	}
	/* value(s) loop until there is no "next" ... */
	while (value != NULL) {
		object_value.object_property = rpm_property->propertyIdentifier;
		object_value.array_index = rpm_property->propertyArrayIndex;
		object_value.value = value;
		switch (rpm_property->propertyIdentifier) {
				/* These are all arrays, so they open and close with braces */
			case PROP_OBJECT_LIST:
			case PROP_STATE_TEXT:
			case PROP_STRUCTURED_OBJECT_LIST:
			case PROP_SUBORDINATE_ANNOTATIONS:
			case PROP_SUBORDINATE_LIST:
				if (false/*Using_Walked_List*/) {
					if ((rpm_property->propertyArrayIndex == 0) &&
						(value->tag == BACNET_APPLICATION_TAG_UNSIGNED_INT)) {
						/* Grab the value of the Object List length - don't
						 * print it! */
						Walked_List_Length = value->type.Unsigned_Int;
						if (rpm_property->propertyIdentifier ==
							PROP_OBJECT_LIST) {
							Object_List_Length = value->type.Unsigned_Int;
						}
						break;
					} else {
						assert(
							Walked_List_Index ==
							(uint32_t)rpm_property->propertyArrayIndex);
					}
				} else {
					Walked_List_Index++;
					/* If we got the whole Object List array in one RP call,
					 * keep the Index and List_Length in sync as we cycle
					 * through. */
					if (rpm_property->propertyIdentifier == PROP_OBJECT_LIST) {
						Object_List_Length = ++Object_List_Index;
					}
				}
				if (Walked_List_Index == 1) {
					/* If the array is empty (nothing for this first entry),
					 * Make it VTS3-friendly and don't show "Null" as a value.
					 */
					if (value->tag == BACNET_APPLICATION_TAG_NULL) {
						result=0;//TODO: NULL value
						break;
					}

					/* Open this Array of Objects for the first entry (unless
					 * opening brace has already printed, since this is an array
					 * of values[] ) */
				}

				if (rpm_property->propertyIdentifier == PROP_OBJECT_LIST) {
					if (value->tag != BACNET_APPLICATION_TAG_OBJECT_ID) {
						break;
					}
					/* Store the object list so we can interrogate
					   each object. */
					object_list_element = KEY_ENCODE(
						value->type.Object_Id.type,
						value->type.Object_Id.instance);
					/* We don't have anything to put in the data pointer
					 * yet, so just leave it null.  The key is Key here. */
					Keylist_Data_Add(Object_List, object_list_element, NULL);
				} else if (
					rpm_property->propertyIdentifier == PROP_STATE_TEXT) {
					/* Make sure it fits within 31 chars for original VTS3
					 * limitation. If longer, take first 15 dash, and last 15
					 * chars. */
					if (value->type.Character_String.length > 31) {
						int iLast15idx =
							value->type.Character_String.length - 15;
						value->type.Character_String.value[15] = '-';
						memmove(
							&value->type.Character_String.value[16],
							&value->type.Character_String.value[iLast15idx],
							15);
						value->type.Character_String.value[31] = 0;
						value->type.Character_String.length = 31;
					}
				} else if (
					rpm_property->propertyIdentifier == PROP_SUBORDINATE_LIST) {
					if (value->tag !=
						BACNET_APPLICATION_TAG_DEVICE_OBJECT_REFERENCE) {
						break;
					}
				}
				convertM(&object_value, result);
				//cout << __func__ << ": found result="<<result<<endl;
				if ((Walked_List_Index < Walked_List_Length) ||
					(value->next != NULL)) {
					/* There are more. */
				} else {
				}
				break;

			case PROP_PROTOCOL_OBJECT_TYPES_SUPPORTED:
			case PROP_PROTOCOL_SERVICES_SUPPORTED:
				break;

				/* Our special non-existent case; do nothing further here. */
			case PROP_PROTOCOL_CONFORMANCE_CLASS:
				break;

			default:
				/* First, if this is a date type, it needs a different format
				 * for VTS, so pretty print it. */
				if (true &&
					(object_value.value->tag == BACNET_APPLICATION_TAG_DATE)) {
					/* This would be PROP_LOCAL_DATE, or OBJECT_DATETIME_VALUE,
					 * or OBJECT_DATE_VALUE					 */
				} else {
					/* Some properties are presented just as '?' in an EPICS;
					 * screen these out here, unless ShowValues is true.  */
					switch (rpm_property->propertyIdentifier) {
						case PROP_DEVICE_ADDRESS_BINDING:
							/* Make it VTS3-friendly and don't show "Null"
							 * as a value. */
							if (value->tag == BACNET_APPLICATION_TAG_NULL) {
								//fprintf(stdout, "?");
								break;
							}
							BACNET_STACK_FALLTHROUGH();
						case PROP_DAYLIGHT_SAVINGS_STATUS:
						case PROP_LOCAL_TIME:
						case PROP_LOCAL_DATE: /* Only if !ShowValues */
						case PROP_PRESENT_VALUE:
						case PROP_PRIORITY_ARRAY:
						case PROP_RELIABILITY:
						case PROP_UTC_OFFSET:
						case PROP_DATABASE_REVISION:
							BACNET_STACK_FALLTHROUGH();
						default:
							convertM(&object_value, result);
							//cout << __func__ << ": found default result="<<result<<endl;
							//bacapp_print_value(stdout, &object_value);
							break;
					}
				}
				if (value->next != NULL) {
					/* there's more! */
				} else {
					if (property_list_writable_member(
							object_type, rpm_property->propertyIdentifier)) {
					}
				}
				break;
		}

		old_value = value;
		value = value->next; /* next or NULL */
		free(old_value);
	} /* End while loop */
}

//+------------------------------------------------------------------
//
//	method:			readthread::readthread()
//
//	description:	readthread constructor
//
//-------------------------------------------------------------------
//readthread::readthread(Tango::DeviceImpl* devImpl):Tango::LogAdapter(devImpl),cond(&mut)
readthread::readthread(Tango::DeviceImpl* devImpl):Tango::LogAdapter(devImpl)
{ 
	DEBUG_STREAM << "readthread::readthread(): constructor..." << endl;
	device = devImpl;
	waitflag = true;

}

//+------------------------------------------------------------------
//
//	method:			readthread::~readthread()
//
//	description:	readthread destructor
//
//-------------------------------------------------------------------
readthread::~readthread()
{
	DEBUG_STREAM << "readthread::~readthread(): destructor... !" << endl;
	cleanup();
	DEBUG_STREAM << "readthread::~readthread(): exiting... !" << endl;
}

//+------------------------------------------------------------------
//
//	method:			readthread::run()
//
//	description:	Run
//
//-------------------------------------------------------------------
void readthread::run(void *)
{
	DEBUG_STREAM << "readthread::run(): running... !" << endl;
	/*int pausesec,pausenano,pausesec_dfl,pausenano_dfl,
		cnt_err = 0,
		max_cnt_err = 3,
		ret;*/

	BacnetConf *_device = static_cast<BacnetConf *>(device);
	abortflag = false;
	while(!_device->created_attr && !abortflag)
	{
		sleep(1);
		DEBUG_STREAM << "readthread::run(): waiting for attr to be created..." << endl;
	}
	/* setup my info */
	Device_Set_Object_Instance_Number(BACNET_MAX_INSTANCE);
	address_init();
	Init_Service_Handlers();
	dlenv_init();

	BACNET_ADDRESS src = { 0 }; /* address where message came from */
	uint16_t pdu_len = 0;
	unsigned timeout = 100; /* milliseconds */
	unsigned max_apdu = 0;
	int tag_value_arg = 0;
	time_t elapsed_seconds = 0;
	time_t last_seconds = 0;
	time_t current_seconds = 0;
	time_t timeout_seconds = 0;
	timeval last_read;
	gettimeofday(&last_read,NULL);
	timeval last_sleep;
	gettimeofday(&last_sleep,NULL);
	timeval now;
	int32_t num_reading=0;
	uint8_t buffer[MAX_PDU] = { 0 };

	char *property_token = NULL;
	unsigned property_id = 0;
	unsigned property_array_index = 0;
	int scan_count = 0;
	unsigned object_type = 0;
	unsigned int target_args = 0;
	bool status = false;

	dev_map_t dev_map;
	attr_map_t::iterator it_attr;
	dev_map_t::iterator it_dev;
	for(it_attr=_device->attr_map.begin();it_attr!=_device->attr_map.end(); it_attr++)
	{
		object_t tmp;
		tmp.s_object_instance =it_attr->second.s_object_instance;
		tmp.object_instance =it_attr->second.object_instance;
		tmp.s_object_type =it_attr->second.s_object_type;
		tmp.object_type =it_attr->second.object_type;
		tmp.s_property = it_attr->second.s_property;
		tmp.property = it_attr->second.property;
		tmp.property_index = it_attr->second.property_index;
		std::string s_device_instance = it_attr->second.s_device_instance;
		it_dev = dev_map.find(s_device_instance);
		if(it_dev == dev_map.end())
		{
			vector<object_t> v_tmp;
			v_tmp.push_back(tmp);
			dev_map.insert(make_pair(s_device_instance,v_tmp));
		}
		else
		{
			it_dev->second.push_back(tmp);
		}
	}

	for(it_dev=dev_map.begin();it_dev!=dev_map.end(); it_dev++)
	{
		uint32_t Target_Device_Object_Instance = BACNET_MAX_INSTANCE;
		BACNET_READ_ACCESS_DATA *rpm_object = NULL;
		BACNET_PROPERTY_REFERENCE *rpm_property = NULL;
		Target_Device_Object_Instance = strtol(it_dev->first.c_str(), NULL, 0);
		vector<object_t>::iterator it_obj;
		read_data_t tmp;
		tmp.Request_Invoke_ID = 0;
		tmp.found = false;
		tmp.processed = false;
		tmp.first = true;
		memset(&tmp.Read_Property_Multiple_Data,0,sizeof(BACnet_RPM_Service_Data_t));
		tmp.Read_Property_Multiple_Data.new_data = false;
		gettimeofday(&tmp.last_read, NULL);
		//tmp.Target_Address;
		int num_obj=0;
		for(it_obj = it_dev->second.begin(); (it_obj != it_dev->second.end()); it_obj++)
		{
			if (rpm_object) {
				rpm_object->next =
					static_cast<BACNET_READ_ACCESS_DATA*>(calloc(1, sizeof(BACNET_READ_ACCESS_DATA)));
				rpm_object = rpm_object->next;
			} else {
				tmp.Read_Access_Data =
					static_cast<BACNET_READ_ACCESS_DATA*>(calloc(1, sizeof(BACNET_READ_ACCESS_DATA)));
				rpm_object = tmp.Read_Access_Data;
				//atexit(cleanup);
			}
			rpm_object->object_type = static_cast<BACNET_OBJECT_TYPE>(it_obj->object_type);
			rpm_object->object_instance = it_obj->object_instance;

			rpm_property = static_cast<BACNET_PROPERTY_REFERENCE*>(calloc(1, sizeof(BACNET_PROPERTY_REFERENCE)));
			rpm_object->listOfProperties = rpm_property;
			/* add all the properties and optional index to our list */
			while (rpm_property)
			{
				rpm_property->propertyIdentifier = static_cast<BACNET_PROPERTY_ID>(it_obj->property);
				if (it_obj->property_index != -1) {
					rpm_property->propertyArrayIndex = it_obj->property_index;
				} else {
					rpm_property->propertyArrayIndex = BACNET_ARRAY_ALL;
				}
				/* is there another property? NO, not supported*/
				{
					rpm_property->next = NULL;
					break;
				}
			}
			if(++num_obj>=_device->maxReadPropertyMultiple)
			{
				read_data_map_t::iterator it_tmp= read_access_data_map.find(Target_Device_Object_Instance);
				if(it_tmp == read_access_data_map.end())
				{
					vector<read_data_t> v_tmp;
					v_tmp.push_back(tmp);
					read_access_data_map.insert(make_pair(Target_Device_Object_Instance,v_tmp));
				}
				else
				{
					it_tmp->second.push_back(tmp);
				}
				num_obj = 0;
				rpm_object = NULL;
				rpm_property = NULL;
				tmp.Read_Access_Data = NULL;
				memset(&tmp.Read_Property_Multiple_Data,0,sizeof(BACnet_RPM_Service_Data_t));
				tmp.Read_Property_Multiple_Data.new_data = false;
			}
		}
		read_data_map_t::iterator it_tmp= read_access_data_map.find(Target_Device_Object_Instance);
		if(it_tmp == read_access_data_map.end())
		{
			vector<read_data_t> v_tmp;
			v_tmp.push_back(tmp);
			read_access_data_map.insert(make_pair(Target_Device_Object_Instance,v_tmp));
		}
		else
		{
			it_tmp->second.push_back(tmp);
		}
	}

	atexit(datalink_cleanup);
	/* configure the timeout values */
	last_seconds = time(NULL);
	timeout_seconds = (apdu_timeout() / 1000) * apdu_retries();
	/* try to bind with the device */
	for(read_data_map_t::iterator it_rd = read_access_data_map.begin(); it_rd != read_access_data_map.end(); it_rd++)
	{
		BACNET_ADDRESS Target_Address;
		bool found = address_bind_request(
			/*Target_Device_Object_Instance*/it_rd->first, &max_apdu, &Target_Address);
		if (!found) {
			Send_WhoIs(
				it_rd->first/*Target_Device_Object_Instance*/, it_rd->first/*Target_Device_Object_Instance*/);
		}
		vector<read_data_t>::iterator it_rd2;
		for(it_rd2=it_rd->second.begin(); it_rd2 != it_rd->second.end(); it_rd2++)
		{
			it_rd2->found = found;
			memcpy(&it_rd2->Target_Address, &Target_Address, sizeof(Target_Address));
		}
	}
	/* loop forever */
	while (!abortflag)
	{
		/* increment timer - exit if timed out */
		current_seconds = time(NULL);
		gettimeofday(&now, NULL);
		double tdiff_read = dtdiff(now,last_read);
		double tdiff_sleep = dtdiff(now, last_sleep);
		double time_to_sleep = 1.0;
		if(tdiff_sleep > 1.0)
			time_to_sleep = 1.0;
		else
			time_to_sleep= 1.0 - tdiff_sleep;
		DEBUG_STREAM << __func__ << ": current_seconds=" << current_seconds << " last_seconds="<<last_seconds << " tdiff_read="<<tdiff_read<<" tdiff_sleep=" << tdiff_sleep << " time_to_sleep="<<time_to_sleep;
		if(tdiff_read < _device->pollingPeriod && time_to_sleep > 0.0)
		{
			DEBUG_STREAM << __func__ << ": going to sleep " << time_to_sleep << " s";
			abort_sleep(time_to_sleep);
			gettimeofday(&last_sleep, NULL);
		}
		else
		{
			abort_sleep(0.005);
		}


		/* at least one second has passed */
		if (current_seconds != last_seconds) {
			tsm_timer_milliseconds(((current_seconds - last_seconds) * 1000));
			datalink_maintenance_timer(current_seconds - last_seconds);
		}
		/*if (Error_Detected) {
			INFO_STREAM << __func__ << ": Error_Detected in while (!abortflag) loop" << endl;
			break;
		}*/
		for(read_data_map_t::iterator it_rd = read_access_data_map.begin(); it_rd != read_access_data_map.end(); it_rd++)
		{
			vector<read_data_t>::iterator it_rd2b=it_rd->second.begin();
			/* wait until the device is bound, or timeout and quit */
			if (!it_rd2b->found) {
				DEBUG_STREAM << "readthread::run(): dev="<<(int)it_rd->first<<" before address_bind_request" << endl;
				it_rd2b->found = address_bind_request(
					it_rd->first/*Target_Device_Object_Instance*/, &max_apdu, &it_rd2b->Target_Address);
				//DEBUG_STREAM << "readthread::run(): dev="<<(int)it_rd->first<<" binded to " <<  print_bacnet_address(&it_rd2b->Target_Address) << " found="<<(int)it_rd2b->found;
			}
			if (it_rd2b->found)
			{
				//DEBUG_STREAM << "readthread::run(): dev="<<(int)it_rd->first<<" binded to " <<  print_bacnet_address(&it_rd2b->Target_Address) << " now looping read_access_data_map second size=" <<  it_rd->second.size();
				vector<read_data_t>::iterator it_rd2;
				for(it_rd2=it_rd->second.begin(); it_rd2 != it_rd->second.end(); it_rd2++)
				{
					it_rd2->found = it_rd2b->found;
					memcpy(&it_rd2->Target_Address, &it_rd2b->Target_Address, sizeof(it_rd2b->Target_Address));
					//DEBUG_STREAM << "readthread::run(): dev="<<(int)it_rd->first<<" copyed target address=" <<  print_bacnet_address(&it_rd2->Target_Address);

					double tdiff_read_single = dtdiff(now,it_rd2->last_read);
					DEBUG_STREAM << "readthread::run(): dev="<<(int)it_rd->first<<" before Send_Read_Property_Multiple_Request Request_Invoke_ID="<<(int)it_rd2->Request_Invoke_ID << " tdiff_read_single="<<tdiff_read_single<<endl;
					if ((tdiff_read_single > _device->pollingPeriod || it_rd2->first) && (it_rd2->Request_Invoke_ID == 0 || (it_rd2->Request_Invoke_ID > 0 && it_rd2->processed)))
					{
						it_rd2->processed = false;
						it_rd2->first = false;
						it_rd2->Request_Invoke_ID = Send_Read_Property_Multiple_Request(
							&buffer[0], sizeof(buffer), it_rd->first/*Target_Device_Object_Instance*/,
							it_rd2->Read_Access_Data);
						if (it_rd2->Request_Invoke_ID == 0) {
							INFO_STREAM << __func__ << ": dev="<<(int)it_rd->first<<" Error: failed to send request!";
							abort_sleep(1);
							break;
						}
						abort_sleep(0.002);
						num_reading++;
						DEBUG_STREAM << "readthread::run(): dev="<<(int)it_rd->first<<" after Send_Read_Property_Multiple_Request Request_Invoke_ID="<<(int)it_rd2->Request_Invoke_ID << endl;
					}
				}
			}
			else
			{
				/* increment timer - exit if timed out */
				elapsed_seconds += (current_seconds - last_seconds);
				if (elapsed_seconds > timeout_seconds) {
					INFO_STREAM << __func__ << ": dev="<<(int)it_rd->first<<" Error: APDU Timeout!" << endl;
					vector<read_data_t>::iterator it_rd2;
					for(it_rd2=it_rd->second.begin(); it_rd2 != it_rd->second.end(); it_rd2++)
					{
						it_rd2->Error_Detected = true;
					}
					break;
				}
			}

			/* returns 0 bytes on timeout */
			pdu_len = datalink_receive(&src, &Rx_Buf[0], MAX_MPDU, timeout);
			DEBUG_STREAM << "readthread::run(): dev="<<(int)it_rd->first<<" datalink_receive -> pdu_len="<<pdu_len;

			/* process */
			if (pdu_len) {
				npdu_handler(&src, &Rx_Buf[0], pdu_len);
			}
			else
			{
				DEBUG_STREAM << __func__ << ": dev="<<(int)it_rd->first<<" read " << pdu_len << " bytes";
			}

			vector<read_data_t>::iterator it_rd2;
			for(it_rd2=it_rd->second.begin(); it_rd2 != it_rd->second.end() && it_rd2->found; it_rd2++)
			{

				if ((it_rd2->Read_Property_Multiple_Data.new_data) &&
							(it_rd2->Request_Invoke_ID ==
							it_rd2->Read_Property_Multiple_Data.service_data.invoke_id))
				{
					it_rd2->Read_Property_Multiple_Data.new_data = false;
					it_rd2->processed = true;
					gettimeofday(&it_rd2->last_read, NULL);
					DEBUG_STREAM << __func__ << ": dev="<<(int)it_rd->first<<" NEW DATA num_reading=" << num_reading;
					num_reading = (num_reading > 1) ? num_reading-1 : 0;
					if(num_reading == 0)
						gettimeofday(&last_read, NULL);

					BACNET_READ_ACCESS_DATA *old_rpm_data;
					BACNET_PROPERTY_REFERENCE *rpm_property;
					BACNET_PROPERTY_REFERENCE *old_rpm_property;
					BACNET_APPLICATION_DATA_VALUE *value;
					BACNET_APPLICATION_DATA_VALUE *old_value;
					bool bHasObjectList = false;
					bool bHasStructuredViewList = false;
					BACNET_READ_ACCESS_DATA *rpm_data=it_rd2->Read_Property_Multiple_Data.rpm_data;
					while (rpm_data) {
						rpm_property = rpm_data->listOfProperties;
						while (rpm_property) {
							/* For the GET_LIST_OF_ALL_RESPONSE case,
							* just keep what property this was */
							attr_map_t::iterator it_att;
							for(it_attr=_device->attr_map.begin(); it_attr != _device->attr_map.end(); it_attr++)
							{
								if(it_attr->second.device_instance == it_rd->first &&
								it_attr->second.object_type == rpm_data->object_type &&
								it_attr->second.object_instance == rpm_data->object_instance &&
								it_attr->second.property == rpm_property->propertyIdentifier)
								{
									break;
								}
							}
							if(it_attr == _device->attr_map.end())
							{
								INFO_STREAM << __func__ << ": NOT found attribute for device_instance="<<it_rd->first<<" object_type="<<rpm_data->object_type
									<<" object_instance="<<rpm_data->object_instance<<" property_id="<<rpm_property->propertyIdentifier;
								old_rpm_property = rpm_property;
								rpm_property = rpm_property->next;
								free(old_rpm_property);
								continue;
							}
							if(it_attr->second.type == string(BAC_FLOAT))
							{
								float result;
								ConvertPropertyData(
									rpm_data->object_type, rpm_data->object_instance,
									rpm_property,
									result);
								it_attr->second.value_f = result;
								_device->push_change_event(it_attr->second.name,&it_attr->second.value_f,/*NULL,Tango::ATTR_VALID,*/ 1/*size*/, 0, false);
								_device->push_archive_event(it_attr->second.name,&it_attr->second.value_f,/*NULL,Tango::ATTR_VALID,*/ 1/*size*/, 0, false);
								DEBUG_STREAM << __func__ << ": name="<<it_attr->second.name<<" DeviceInstance="<< (int)it_rd->first<<" type="<<bactext_object_type_name(rpm_data->object_type)<<" #"<< rpm_data->object_instance << " property=" << bactext_property_name(rpm_property->propertyIdentifier) <<"="<<result << endl;
							}
							else if(it_attr->second.type == string(BAC_DOUBLE))
							{
								double result;
								ConvertPropertyData(
									rpm_data->object_type, rpm_data->object_instance,
									rpm_property,
									result);
								it_attr->second.value_d = result;
								_device->push_change_event(it_attr->second.name,&it_attr->second.value_d,/*NULL,Tango::ATTR_VALID,*/ 1/*size*/, 0, false);
								_device->push_archive_event(it_attr->second.name,&it_attr->second.value_d,/*NULL,Tango::ATTR_VALID,*/ 1/*size*/, 0, false);
								DEBUG_STREAM << __func__ << ": name="<<it_attr->second.name<<" DeviceInstance="<< (int)it_rd->first<<" type="<<bactext_object_type_name(rpm_data->object_type)<<" #"<< rpm_data->object_instance << " property=" << bactext_property_name(rpm_property->propertyIdentifier) <<"="<<result << endl;
							}
							else if(it_attr->second.type == string(BAC_INT32))
							{
								Tango::DevLong result;
								ConvertPropertyData(
									rpm_data->object_type, rpm_data->object_instance,
									rpm_property,
									result);
								it_attr->second.value_l = result;
								_device->push_change_event(it_attr->second.name,&it_attr->second.value_l,/*NULL,Tango::ATTR_VALID,*/ 1/*size*/, 0, false);
								_device->push_archive_event(it_attr->second.name,&it_attr->second.value_l,/*NULL,Tango::ATTR_VALID,*/ 1/*size*/, 0, false);
								DEBUG_STREAM << __func__ << ": name="<<it_attr->second.name<<" DeviceInstance="<< (int)it_rd->first<<" type="<<bactext_object_type_name(rpm_data->object_type)<<" #"<< rpm_data->object_instance << " property=" << bactext_property_name(rpm_property->propertyIdentifier) <<"="<<result << endl;
							}
							else if(it_attr->second.type == string(BAC_INT16))
							{
								Tango::DevShort result;
								ConvertPropertyData(
									rpm_data->object_type, rpm_data->object_instance,
									rpm_property,
									result);
								it_attr->second.value_s = result;
								_device->push_change_event(it_attr->second.name,&it_attr->second.value_s,/*NULL,Tango::ATTR_VALID,*/ 1/*size*/, 0, false);
								_device->push_archive_event(it_attr->second.name,&it_attr->second.value_s,/*NULL,Tango::ATTR_VALID,*/ 1/*size*/, 0, false);
								DEBUG_STREAM << __func__ << ": name="<<it_attr->second.name<<" DeviceInstance="<< (int)it_rd->first<<" type="<<bactext_object_type_name(rpm_data->object_type)<<" #"<< rpm_data->object_instance << " property=" << bactext_property_name(rpm_property->propertyIdentifier) <<"="<<result << endl;
							}
							else if(it_attr->second.type == string(BAC_BYTE))
							{
								Tango::DevUChar result;
								ConvertPropertyData(
									rpm_data->object_type, rpm_data->object_instance,
									rpm_property,
									result);
								it_attr->second.value_uc = result;
								_device->push_change_event(it_attr->second.name,&it_attr->second.value_uc,/*NULL,Tango::ATTR_VALID,*/ 1/*size*/, 0, false);
								_device->push_archive_event(it_attr->second.name,&it_attr->second.value_uc,/*NULL,Tango::ATTR_VALID,*/ 1/*size*/, 0, false);
								DEBUG_STREAM << __func__ << ": name="<<it_attr->second.name<<" DeviceInstance="<< (int)it_rd->first<<" type="<<bactext_object_type_name(rpm_data->object_type)<<" #"<< rpm_data->object_instance << " property=" << bactext_property_name(rpm_property->propertyIdentifier) <<"="<<result << endl;
							}
							else if(it_attr->second.type == string(BAC_BOOL))
							{
								Tango::DevBoolean result;
								ConvertPropertyData(
									rpm_data->object_type, rpm_data->object_instance,
									rpm_property,
									result);
								it_attr->second.value_b = result;
								_device->push_change_event(it_attr->second.name,&it_attr->second.value_b,/*NULL,Tango::ATTR_VALID,*/ 1/*size*/, 0, false);
								_device->push_archive_event(it_attr->second.name,&it_attr->second.value_b,/*NULL,Tango::ATTR_VALID,*/ 1/*size*/, 0, false);
								DEBUG_STREAM << __func__ << ": name="<<it_attr->second.name<<" DeviceInstance="<< (int)it_rd->first<<" type="<<bactext_object_type_name(rpm_data->object_type)<<" #"<< rpm_data->object_instance << " property=" << bactext_property_name(rpm_property->propertyIdentifier) <<"="<<result << endl;
							}
							old_rpm_property = rpm_property;
							rpm_property = rpm_property->next;
							free(old_rpm_property);
						}
						old_rpm_data = rpm_data;
						rpm_data = rpm_data->next;
						free(old_rpm_data);
					}
					if (tsm_invoke_id_free(it_rd2->Request_Invoke_ID))
					{
						INFO_STREAM << __func__ << ": dev="<<(int)it_rd->first<<" tsm_invoke_id_free: " << (int)it_rd2->Request_Invoke_ID << " is free" << endl;
						it_rd2->Request_Invoke_ID = 0;
					}
					else
					{
						//assert(false); /* How can this be? */
						INFO_STREAM << __func__ << ": dev="<<(int)it_rd->first<<" tsm_invoke_id_free: " << (int)it_rd2->Request_Invoke_ID << " is NOT free (How can this be?)" << endl;
						if (tsm_invoke_id_failed(it_rd2->Request_Invoke_ID))
						{
							//fprintf(stderr, "\rError: TSM Timeout!\n");
							INFO_STREAM << __func__ << ": dev="<<(int)it_rd->first<<" Error: TSM Timeout!" << endl;
							tsm_free_invoke_id(it_rd2->Request_Invoke_ID);
							//Error_Detected = true;
							/* try again or abort? */
							//break;
							abort_sleep(1);//TODO: configure period
							it_rd2->Request_Invoke_ID = 0;
						}
					}
					it_rd2->Request_Invoke_ID = 0;//TODO: add GZS
				}
			}
		}//end looping read_access_data_map

		/* keep track of time for next check */
		last_seconds = current_seconds;
	}
	DEBUG_STREAM << "readthread::run(): exit!!!" << endl;
}

//+------------------------------------------------------------------
//
//	method:			readthread::abort_sleep
//
//	description:	Resume from sleep if abort_flag set (sec.)
//
//-------------------------------------------------------------------
void readthread::abort_sleep(double time)
{
	omni_mutex_lock sync(*this);
	long time_millis = (long)(time*1000);
	int res = wait(time_millis);
}


}


